#!/usr/bin/env python2

# pynfs - Python NFS4 tools
#
# Written by Peter strand <peter@cendio.se>
# Copyright (C) 2001 Cendio Systems AB (http://www.cendio.se)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License. 
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# TODO: Clean up. Support TCP. More error checks. 

import sys, os
import nfs4lib
from nfs4constants import *
from nfs4types import *

# FILEID added to work around server bug if do readdir without any attributes set
# RDATTR added to work around server returning _WRONGSEC during create_leading_paths()
anyattr = nfs4lib.list2attrmask([FATTR4_FILEID, FATTR4_RDATTR_ERROR])

def clean_dir(directory):
    fh = ncl.do_getfh(directory)
    entries = ncl.do_readdir(fh,anyattr)
    names = [entry.name for entry in entries]
    
    stateid = stateid4(0, "")
    obj_attributes = nfs4lib.dict2fattr({FATTR4_MODE: 0755})

    for name in names:
        # We separate setattr and remove to avoid an inode locking bug
        lookup_dir_ops = ncl.lookup_path(directory + [name])
        operations = [ncl.putrootfh_op()] + lookup_dir_ops
        operations.append(ncl.setattr_op(stateid, obj_attributes))
        res = ncl.compound(operations)
        if res.status != NFS4_OK:
            raise "Cannot clean directory %s, error %s" % (directory, nfs4lib.nfsstat4_id[res.status])

        lookup_dir_ops = ncl.lookup_path(directory)
        operations = [ncl.putrootfh_op()] + lookup_dir_ops
        operations.append(ncl.remove_op(name))
        res = ncl.compound(operations)

        if res.status == NFS4ERR_NOTEMPTY:
            # Recursive deletion
            clean_dir(directory + [name])
            # Remove dir itself
            operations = [ncl.putrootfh_op()] + lookup_dir_ops
            operations.append(ncl.remove_op(name))
            res = ncl.compound(operations)
            nfs4lib.check_result(res)
        if res.status != NFS4_OK:
            raise "Cannot clean directory %s, error %s" % (directory, nfs4lib.nfsstat4_id[res.status])

    # Verify that all files were removed
    entries = ncl.do_readdir(fh, anyattr)
    if entries:
        raise "Cannot clean directory %s" % directory

def create_dir(ncl, curdir, create_dir):
    operations = [ncl.putrootfh_op()] + ncl.lookup_path(curdir)
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE:0755})
    createop = ncl.create_op(objtype, create_dir, attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

def create_leading_paths(ncl, prefix):
    for x in range(len(prefix)):
        leading_comps = prefix[0:x+1]
        curdir = leading_comps[:-1]
        lastcomp = leading_comps[-1]
        fh = ncl.do_getfh(curdir)
        entries = ncl.do_readdir(fh, anyattr)
        names = [entry.name for entry in entries]
        if not lastcomp in names:
            create_dir(ncl, curdir, lastcomp)

def main(ncl, unix_prefix):
    ncl.init_connection()
    prefix = nfs4lib.unixpath2comps(unix_prefix)
    putrootfhop = ncl.putrootfh_op()

    lookup_treeroot = [putrootfhop] + ncl.lookup_path(prefix)
    lookup_dev = lookup_treeroot + ncl.lookup_path(["dev"])
    lookup_doc = lookup_treeroot + ncl.lookup_path(["doc"])
    lookup_src = lookup_treeroot + ncl.lookup_path(["src"])
    lookup_tmp = lookup_treeroot + ncl.lookup_path(["tmp"])
    lookup_private = lookup_treeroot + ncl.lookup_path(["private"])

    print "Creating path", unix_prefix, prefix
    create_leading_paths(ncl, prefix)

    print "Clearing", unix_prefix
    clean_dir(prefix)

    print "Creating /dev"
    operations = lookup_treeroot[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755})
    createop = ncl.create_op(objtype, "dev", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating /dev/floppy"
    operations = lookup_dev[:]
    objtype = createtype4(type=NF4LNK, linkdata="fd0")
#    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755})
    attrs = nfs4lib.dict2fattr({})
    createop = ncl.create_op(objtype, "floppy", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    try:
        nfs4lib.check_result(res)
    except nfs4lib.BadCompoundRes, e:
        print "Could not create /dev/floppy"
        print e

    print "Creating /dev/fd0"
    operations = lookup_dev[:]
    devdata = specdata4(2, 0)
    objtype = createtype4(type=NF4BLK, devdata=devdata)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
    createop = ncl.create_op(objtype, "fd0", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    try:
        nfs4lib.check_result(res)
    except nfs4lib.BadCompoundRes, e:
        print "Could not create /dev/fd0"
        print e

    print "Creating /dev/ttyS0"
    operations = lookup_dev[:]
    devdata = specdata4(4, 64)
    objtype = createtype4(type=NF4CHR, devdata=devdata)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
    createop = ncl.create_op(objtype, "ttyS0", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    try:
        nfs4lib.check_result(res)
    except nfs4lib.BadCompoundRes, e:
        print "Could not create /dev/ttyS0"
        print e

    print "Creating /dev/log"
    operations = lookup_dev[:]
    objtype = createtype4(type=NF4SOCK)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
    createop = ncl.create_op(objtype, "log", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    try:
        nfs4lib.check_result(res)
    except nfs4lib.BadCompoundRes, e:
        print "Could not create /dev/log"
        print e

    print "Creating /dev/initctl"
    operations = lookup_dev[:]
    objtype = createtype4(type=NF4FIFO)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0644 })
    createop = ncl.create_op(objtype, "initctl", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    try:
        nfs4lib.check_result(res)
    except nfs4lib.BadCompoundRes, e:
        print "Could not create /dev/initctl"
        print e

    print "Creating /doc"
    operations = lookup_treeroot[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    createop = ncl.create_op(objtype, "doc", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating /doc/README"
    remote = nfs4lib.NFS4OpenFile(ncl)
    remote.open(unix_prefix + "/doc/README", "w")
    #remote.open(os.path.join(unix_prefix, "doc/README"), "w")
    data ="""\
    Welcome to this NFS4 server.
    Enjoy.
    """
    remote.write(data)
    remote.close()

    print "Creating directory doc/porting"
    operations = lookup_doc[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    createop = ncl.create_op(objtype, "porting", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating doc/porting/TODO"
    remote = nfs4lib.NFS4OpenFile(ncl)
    remote.open(unix_prefix + "/doc/porting/TODO", "w")
    #remote.open(os.path.join(unix_prefix, "doc/porting/TODO"), "w")
    data ="""\
    Need to work on DNIX support...
    Enjoy.
    """
    remote.write(data)
    remote.close()

    print "Creating src"
    operations = lookup_treeroot[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    createop = ncl.create_op(objtype, "src", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating src/hello.c"
    remote = nfs4lib.NFS4OpenFile(ncl)
    remote.open(unix_prefix + "/src/hello.c", "w")
    #remote.open(os.path.join(unix_prefix, "src/hello.c"), "w")
    data = """\
    #include <stdio.h>
    #include <stdlib.h>

    int main()
    {
        printf("Hello world!\\n");
        exit(0);
    }
    """
    remote.write(data)
    remote.close()

    print "Creating tmp"
    operations = lookup_treeroot[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    createop = ncl.create_op(objtype, "tmp", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    '''
    print "Creating tmp/gazonk"
    operations = lookup_tmp[:]
    objtype = createtype4(type=NF4DIR)
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    createop = ncl.create_op(objtype, "gazonk", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating tmp/gazonk/foo.c"
    remote = nfs4lib.NFS4OpenFile(ncl)
    remote.open(unix_prefix + "/tmp/gazonk/foo.c", "w")
    #remote.open(os.path.join(unix_prefix, "tmp/gazonk/foo.c"), "w")
    data = """\
    #include <stdio.h>
    #include <stdlib.h>

    int main()
    {
        printf("Hello world!\\n");
        exit(0);
    }
    """
    remote.write(data)
    remote.close()
    '''
    
    print "Creating directory private"
    operations = lookup_treeroot[:]
    attrs = nfs4lib.dict2fattr({FATTR4_MODE: 0755 })
    objtype = createtype4(type=NF4DIR)
    createop = ncl.create_op(objtype, "private", attrs)
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)

    print "Creating private/info.txt"
    remote = nfs4lib.NFS4OpenFile(ncl)
    remote.open(unix_prefix + "/private/info.txt", "w")
    #remote.open(os.path.join(unix_prefix, "private/info.txt"), "w")
    remote.write("Personal data.\n")
    remote.close()

    print "Changing UNIX permissions of private dir to 0000"
    operations = lookup_private[:]
    stateid = stateid4(0, "")
    obj_attributes = nfs4lib.dict2fattr({FATTR4_MODE: 0})
    operations.append(ncl.setattr_op(stateid, obj_attributes))

    res = ncl.compound(operations)
    if res.status == NFS4ERR_ATTRNOTSUPP:
        print "UNIX mode attribute not supported"
    else:
        nfs4lib.check_result(res)

    print "Creating symlink src/doc -> ../doc"
    operations = lookup_src[:]
    objtype = createtype4(type=NF4LNK, linkdata="../doc")
    createop = ncl.create(objtype, "doc")
    operations.append(createop)
    res = ncl.compound(operations)
    nfs4lib.check_result(res)


def usage():
    print """\
Usage: %s [nfs://]<server>[:[port]<prefix>]
 options:
  --uid=<uid>                  Use custom <uid> with AUTH_UNIX security
  --gid=<gid>                  Use custom <gid> with AUTH_UNIX security
  --owner=<owner>              Use custom fattr_owner 
  --owner_group=<owner_group>  Use custom fattr_owner_group
  --sec=<security flavor>      Use given security (default=sys)

Creates tree contents on server for nfs4st testing
Directories and files will be created
under <server>:<prefix>/nfs4st_tree/

<prefix> defaults to /, e.g. the directory nfs4st is
created directly under the server root
""" %  sys.argv[0]
    sys.exit(1)

if __name__ == "__main__":
    
    import getopt

    uid = 0
    gid = 0
    if os.getenv("PYNFS_UID"):
        uid = int(os.getenv("PYNFS_UID"))
    if os.getenv("PYNFS_GID"):
        gid = int(os.getenv("PYNFS_GID"))

    # Reorder arguments, so we can add options at the end 
    ordered_args = []
    for arg in sys.argv[1:]:
        if arg.startswith("-"):
            ordered_args.insert(0, arg)
        else:
            ordered_args.append(arg)

    try:
        options, args = getopt.getopt(ordered_args, '',
                                        ['help','uid=', 'gid=', 
                                        'owner=', 'owner_group=',
                                         'sec='])
    except getopt.error, msg:
        print "\n%s\n" % msg
        usage()

    kwargs = {'uid':uid, 'gid':gid}
    for opt, value in options:
        if opt == "--uid":
            kwargs['uid'] = int(value)
        if opt == "--gid":
            kwargs['gid'] = int(value)
        if opt == "--owner":
            nfs4lib.OWNER = str(value)
        if opt == "--owner_group":
            nfs4lib.OWNER_GROUP = str(value)
        if opt == "--sec":
            import rpc
            try:
                kwargs['sec'] = rpc.supported[value]
            except KeyError:
                print "Unsupported option for --sec : %s" % a
                print "Supported options are:"
                print rpc.supported.keys()
                sys.exit(2)
        if opt == "--help":
            usage()

    if len(args) < 1:
        usage()
    parse_result = nfs4lib.parse_nfs_url(args[0])
    if not parse_result:
        usage()

    (host, portstring, directory) = parse_result

    print host,directory
    
    if portstring:
        port = int(portstring)
    else:
        port = nfs4lib.NFS_PORT

    #   (FRED) - This seems way too unix-centric, but the previous combined
    # use of os.path.join with lots of implicit assumptions that os.sep=='/'
    # was causing problems.
    #   Currently parse_nfs_url forces use of /, so might as well be
    # consistent.  Directory MUST end up unix-like for main to work.
    
    # FIXME - this whole thing, and functions used in nfs4lib,  should be
    # reworked using only python os library.  The nfs4lib functions used
    # (unixpath2comps,  parse_nfs_url), are obviously brittle.  However,
    # need access to windows machine to do properly.

    if not directory:
        directory = "/"

    directory = directory + "/nfs4st_tree"
    #directory = os.path.join(directory, "nfs4st_tree")
    ncl = nfs4lib.create_client(host, port, "tcp", **kwargs)
    main(ncl, directory)

